<?php

namespace Drp\NativeEventApp\Test\Unit\Notifications\Listeners;

use Carbon\Carbon;
use Drp\NativeEventApp\Models\Device;
use Drp\NativeEventApp\Models\Event;
use Drp\NativeEventApp\Models\Notification;
use Drp\NativeEventApp\Models\Registration;
use Drp\NativeEventApp\Models\Segment;
use Drp\NativeEventApp\Notifications\Events\SendNotification;
use Drp\NativeEventApp\Notifications\Events\SendPushNotification;
use Drp\NativeEventApp\Notifications\Listeners\NotificationSender;
use Drp\NativeEventApp\Test\Support\Builders\DeviceBuilder;
use Drp\NativeEventApp\Test\TestCase;
use Illuminate\Support\Facades\Event as EventFacade;

class NotificationSenderTest extends TestCase
{
    /**
     * @var \Drp\NativeEventApp\Notifications\Listeners\NotificationSender
     */
    protected $notificationSender;

    /**
     * @var DeviceBuilder
     */
    protected $deviceBuilder;

    protected function setUp(): void
    {
        parent::setUp();

        $this->notificationSender = app(NotificationSender::class);
        $this->deviceBuilder = app(DeviceBuilder::class);
    }

    /** @test */
    public function it_can_send_notifications_to_all_unique_devices_in_segments(): void
    {
        Carbon::setTestNow('02-04-2020 14:00');

        [$segmentA, $segmentB] = factory(Segment::class)->times(2)->create();

        $this->deviceBuilder
            ->incrementAuthId()
            ->withRegistration(static function (Registration $registration) use ($segmentA): void {
                $registration->joinSegments(collect([$segmentA]));
            })
            ->times(2)
            ->create();

        $this->deviceBuilder
            ->incrementAuthId()
            ->withRegistration(static function (Registration $registration) use ($segmentB): void {
                $registration->joinSegments(collect([$segmentB]));
            })
            ->times(2)
            ->create();

        $notification = Notification::make(factory(Event::class)->create(), 'Test', 'Message')
            ->forSegments(collect([$segmentA]))
            ->sendAt(now())
            ->publish();

        $event = new SendNotification($notification);

        EventFacade::fake();

        $this->notificationSender->handle($event);

        $devices = Device::take(2)->get();

        EventFacade::assertDispatched(SendPushNotification::class, 2);

        foreach ($devices as $device) {
            EventFacade::assertDispatched(
                SendPushNotification::class,
                static function (SendPushNotification $event) use ($device) {
                    return $event->device->id === $device->id;
                }
            );
        }
    }

    /** @test */
    public function it_wont_attempt_to_send_a_notification_to_a_registration_without_a_device(): void
    {
        Carbon::setTestNow('02-04-2020 14:00');

        $segment = factory(Segment::class)->create();

        // Create registration with a device.
        $results = $this->deviceBuilder
            ->withRegistration(static function (Registration $registration) use ($segment): void {
                $registration->joinSegments(collect([$segment]));
            })
            ->create();
        $device = $results->devices->first();

        // Create registration with no device.
        factory(Registration::class)->create([
            'authId' => 2,
        ])->joinSegments(collect([$segment]));

        $this->assertCount(2, Registration::all());
        $this->assertCount(1, Device::all());

        $notification = Notification::make(factory(Event::class)->create(), 'Test', 'Message')
            ->forSegments(collect([$segment]))
            ->sendAt(now())
            ->publish();

        $event = new SendNotification($notification);

        EventFacade::fake();

        $this->notificationSender->handle($event);

        EventFacade::assertDispatched(SendPushNotification::class, 1);

        EventFacade::assertDispatched(
            SendPushNotification::class,
            static function (SendPushNotification $event) use ($device) {
                return $event->device->id === $device->id;
            }
        );
    }
}
